#!/usr/bin/env python3

import datetime
from enum import StrEnum
from pathlib import Path

from fastapi import FastAPI, HTTPException, Response, status
from fastapi.responses import FileResponse, StreamingResponse
from humps import camelize
from pydantic import BaseModel, Field

# Configuration parameters
RECORING_DIR = "/tmp/lohrocorder/recordings/"
JOBS_DIR = "/tmp/lohrocorder/jobs/"


# Helpers
def to_camel(string):
    return camelize(string)


# FastAPI app
app = FastAPI(
    title="LOHROcorder API",
    description="A Python based recording service with API",
)


# Models
class Codec(StrEnum):
    aac = "aac"
    flac = "flac"
    mp3 = "mp3"
    opus = "opus"
    vorbis = "vorbis"
    wav = "wav"


class Quality(StrEnum):
    low = "low"
    medium = "medium"
    high = "high"
    ultra = "ultra"
    lossless = "lossless"


class ProcessingStatusStage(StrEnum):
    unknown = "unknown"
    pending = "pending"
    running = "running"
    finished = "finished"
    failed = "failed"


class CamelBaseModel(BaseModel):
    class Config:
        alias_generator = to_camel
        allow_population_by_field_name = True


class Recording(CamelBaseModel):
    start: list[datetime.datetime]
    end: list[datetime.datetime]
    codec: Codec = Codec.opus

    @property
    def id(self) -> str:
        """
        Returns a unique id for the recording based on the request data.
        Same request data should always result in the same id.
        """
        unique_str = "".join([str(x) for x in self.start + self.end + [self.codec]])
        id = hex(abs(hash(unique_str)))[2:]
        return id

    def filename(self) -> str:
        return f"{self.id}"

    def full_path(self) -> Path:
        return Path(RECORING_DIR) / self.filename()


class RequestForRecording(CamelBaseModel):
    start: list[datetime.datetime]
    end: list[datetime.datetime]
    codec: Codec = Codec.opus


class ProcessingStatus(CamelBaseModel):
    stage: ProcessingStatusStage
    percent: int
    result: str = Field(..., description="The URL of the generated audiofile")
    recording_id: str = Field(..., description="The id of the recording")


# Logic and helpers
def get_unique_recording_id(recording: Recording) -> str:
    """
    Returns a unique id for the recording based on the request data.
    Same request data should always result in the same id.
    """
    unique_str = "".join(
        [str(x) for x in recording.start + recording.end + [recording.codec]]
    )
    id = hex(abs(hash(unique_str)))
    return id


def create_recording_dummy(recording: Recording) -> Path:
    """
    Creates a dummy file for the recording.
    """

    dummy_file = Path("lohrocorder/dummy.mp3")
    export_file = recording.full_path()

    export_file.parent.mkdir(parents=True, exist_ok=True)
    export_file.write_bytes(dummy_file.read_bytes())

    return export_file


# Routes
@app.post(
    "/recordings/", status_code=status.HTTP_202_ACCEPTED, tags=["Export to Audiofile"]
)
async def compose_recording(
    request_for_recording: RequestForRecording,
    response: Response,
) -> ProcessingStatus:
    recording = Recording(
        start=request_for_recording.start,
        end=request_for_recording.end,
        codec=request_for_recording.codec,
    )

    create_recording_dummy(recording)

    result_url = f"/recordings/{recording.id}"

    response.headers["Location"] = result_url

    return ProcessingStatus(
        stage=ProcessingStatusStage.pending,
        percent=10,
        result=result_url,
        recording_id=recording.id,
    )


@app.get(
    "/recordings/{recording_id}",
    response_class=FileResponse,
    responses={404: {"description": "Not found"}},
    response_description="The generated recording file",
    description="Returns the generated recording file",
    tags=["Export to Audiofile"],
)
async def get_recording(recording_id: str) -> FileResponse:
    if not recording_id:
        raise HTTPException(status_code=400, detail="No recording id provided")

    expected_abs_path = Path(RECORING_DIR) / recording_id

    if not expected_abs_path.exists():
        raise HTTPException(status_code=404, detail="Recording not found")

    return FileResponse(expected_abs_path)


@app.get("/recordings/{recording_id}/status", tags=["Export to Audiofile"])
async def get_recording_status(recording_id: str) -> ProcessingStatus:
    return ProcessingStatus(
        recording_id=recording_id,
        stage=ProcessingStatusStage.unknown,
        percent=50,
        result=app.url_path_for("get_recording", recording_id=recording_id),
    )


@app.get("/stream")
async def get_stream(start: datetime.datetime, end: datetime.datetime):
    print(start, end)

    def iterfiles(files_to_stream):
        for file in files_to_stream:
            with file.open("rb") as f:
                yield from f

    test_files = [Path("lohrocorder/dummy.mp3")] * 100

    return StreamingResponse(iterfiles(test_files), media_type="audio/mpeg")


if __name__ == "__main__":
    import uvicorn

    uvicorn.run("lohrocorder.app:app", port=8001, host="0.0.0.0")
