#!/usr/bin/env python3

import datetime
import logging
import pathlib
import tempfile

# Configure Logging
logging.basicConfig(level=logging.DEBUG)


RECORDINGS_ARCHIVE_DIR = pathlib.Path("./recordings-archive").resolve()
RECORDING_NAME_DATETIME_FORMAT = "%Y-%m-%d/%H-%M-%S.mp3"

MAX_ALLOWED_EXPORT_DURATION_DIFFERENCE_S = 2.0


# get all files from recording archieve
def get_recording_archive_files() -> list[pathlib.Path]:
    """
    List all files from the recording archive sorted by name
    """
    archive_files = [
        f for f in list(RECORDINGS_ARCHIVE_DIR.glob("**/*")) if f.is_file()
    ]
    return sorted(archive_files)


# find configuration file
def find_config_file() -> pathlib.Path:
    """
    Find a configuration file for the lohrocorder
    raises FileNotFoundError if no config file was found
    """

    config_filename = "lohrocorder.conf"

    # Search some usual places for the configuration file. Keep the first one
    possible_config_dirs = [
        pathlib.Path.cwd(),
        pathlib.Path("/etc/lohrocorder"),
        pathlib.Path("/etc"),
        pathlib.Path.home() / ".config" / "lohrocorder",
    ]

    for possible_config_dir in possible_config_dirs:
        config_file = possible_config_dir / config_filename
        if config_file.exists():
            logging.info(f"Found config file {config_file}")
            return config_file

    raise FileNotFoundError("Could not find config file")


# Get duration of an audio file using ffprobe
async def get_duration_of_audiofile(audiofile: pathlib.Path) -> float:
    """
    Get the duration of an audio file in seconds

    may raise RuntimeError, TypeError or ValueError
    """

    if not isinstance(audiofile, pathlib.Path):
        error_message = "Audiofile is not a Path object"
        logging.error(error_message)
        raise TypeError(error_message)

    # get length of audio using ffprobe
    proc = await asyncio.create_subprocess_exec(
        "ffprobe",
        "-v",
        "error",
        "-show_entries",
        "format=duration",
        "-of",
        "default=noprint_wrappers=1:nokey=1",
        audiofile,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    stdout, stderr = await proc.communicate()

    # Check if ffprobe was successful and print stderr
    if proc.returncode != 0:
        error_message = (
            f"ffprobe failed with returncode {proc.returncode}: {str(stderr)}"
        )
        logging.error(error_message)
        raise RuntimeError(error_message)

    duration = float(stdout)
    logging.debug(f"Duration of {audiofile}: {duration:0.2f}s")

    return duration


def get_start_of_audiofile(audiofile: pathlib.Path) -> datetime.datetime:
    """
    Get the start datetime of an audio file
    """

    relative_path = audiofile.relative_to(RECORDINGS_ARCHIVE_DIR)
    start = datetime.datetime.strptime(
        str(relative_path), RECORDING_NAME_DATETIME_FORMAT
    )

    return start


async def audiofile_includes_datetime(
    audiofile: pathlib.Path, datetime_of_interest: datetime.datetime
) -> bool:
    """
    Check if the audio file includes the datetime
    """

    audiofile_duration = await get_duration_of_audiofile(audiofile)
    audiofile_start = get_start_of_audiofile(audiofile)
    audiofile_end = audiofile_start + datetime.timedelta(seconds=audiofile_duration)

    return audiofile_start <= datetime_of_interest <= audiofile_end


async def get_recording_file_from_archive(
    datetime_of_interest: datetime.datetime,
) -> pathlib.Path:
    """
    Get the recording file for the datetime from the 24/7 archive
    """

    source_files = get_recording_archive_files()

    # We add an virtual file to use it as a marker. This file does not exist, but
    # if we sort all files including the marker by name, it will be the position of
    # interest. The prior file of the marker is the file we want to check.

    marker_filename = datetime_of_interest.strftime(RECORDING_NAME_DATETIME_FORMAT)
    marker_file = pathlib.Path("./recordings-archive").resolve() / marker_filename

    source_files.append(marker_file)
    source_files.sort()

    recording_file = source_files[source_files.index(marker_file) - 1]

    # Check if the recording file includes the datetime
    if not await audiofile_includes_datetime(recording_file, datetime_of_interest):
        raise FileNotFoundError(
            f"Could not find recording file for {datetime_of_interest}"
        )

    return recording_file


async def get_recording_files_in_range(
    start: datetime.datetime, end: datetime.datetime
):
    audiofile_start = await get_recording_file_from_archive(start)
    audiofile_end = await get_recording_file_from_archive(end)

    source_files = get_recording_archive_files()

    range_start_index = source_files.index(audiofile_start)
    range_end_index = source_files.index(audiofile_end) + 1

    source_files_in_range = source_files[range_start_index:range_end_index]

    logging.debug(f"Found {len(source_files_in_range)} files in range {start} - {end}")

    return source_files_in_range


async def get_duration_sum_of_audiofiles(audiofiles: list[pathlib.Path]) -> float:
    """
    Get the sum of the duration of all audiofiles
    """

    duration_sum = 0.0
    for audiofile in audiofiles:
        duration_sum += await get_duration_of_audiofile(audiofile)

    return duration_sum


# create a recording export using ffmpeg
async def create_recording_export(
    starttime: datetime.datetime,
    endtime: datetime.datetime,
    output_file: pathlib.Path,
    format: str = "mp3",
    bitrate: int = 128,
) -> pathlib.Path:
    # concatenate all audio files in the range using ffmpeg
    audiofiles_in_range = await get_recording_files_in_range(starttime, endtime)

    # Get difference between start of the first audiofile and the starttime we want
    start_offset = starttime - get_start_of_audiofile(audiofiles_in_range[0])
    duration = endtime - starttime

    # TODO: The concept is just a start. We have to allow multiple start- and endtimes.
    # For this we could generate chunks of audiofiles for every pair of start- and
    # endtime. Then we could merge all chunks together to get the final file. This
    # would also allow mulithreading by executing multiple ffmpeg commands in parallel.

    # There is a maximum length for commands executed in a shell. As we expect to merge
    # a lot of files, we need to use a temporary file to store the list of files to
    # merge. ffmpeg is able to handle this file as input.

    with tempfile.NamedTemporaryFile(mode="w") as concat_files_file:
        # create and write list of files for concat
        concat_files_str = "\n".join([f"file '{x}'" for x in audiofiles_in_range])
        concat_files_file.write(concat_files_str)
        concat_files_file.flush()

        ffmpeg_command: list[str] = [
            "ffmpeg",
            "-hide_banner",
            "-loglevel",
            "error",
            "-y",
            "-f",
            "concat",
            "-safe",
            "0",
            "-i",
            concat_files_file.name,
            "-b:a",
            f"{bitrate}k",
            "-ss",
            str(start_offset.total_seconds()),
            "-t",
            str(duration.total_seconds()),
            str(output_file.resolve()),
        ]

        proc = await asyncio.create_subprocess_exec(
            *ffmpeg_command,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        ffmpeg_command_str = " ".join(ffmpeg_command)
        logging.debug(f'execute ffmpeg command: "{ffmpeg_command_str}"')

        _, stderr = await proc.communicate()

    # Check if ffmpeg was successful
    if proc.returncode != 0:
        error_message = (
            f"ffmpeg failed with returncode {proc.returncode}: {str(stderr)}"
        )
        logging.error(error_message)
        raise RuntimeError(error_message)

    # validate output file
    if not output_file.exists():
        error_message = f"ffmpeg failed to create output file {output_file}"
        logging.error(error_message)
        raise RuntimeError(error_message)

    output_file_duration_s = await get_duration_of_audiofile(output_file)
    duration_difference = output_file_duration_s - duration.total_seconds()

    logging.debug(f"Duration difference: {duration_difference:0.2f}s")

    if abs(duration_difference) > MAX_ALLOWED_EXPORT_DURATION_DIFFERENCE_S:
        error_message = f"The export has not the expected duration. Expected: \
            {duration.total_seconds()}s, actual: {output_file_duration_s}s"
        logging.error(error_message)
        raise RuntimeError(error_message)

    logging.info(f"Succesfully created export file {output_file}")

    return output_file


async def main():
    logging.info("Starting lohrocorder processing module in standalone mode")
    config_file = find_config_file()
    if not config_file:
        logging.error("Could not find config file")
        return

    start_datetime = datetime.datetime(2023, 5, 1, 13, 21, 43)
    end_datetime = datetime.datetime(2023, 5, 1, 13, 52, 41)

    await create_recording_export(
        start_datetime, end_datetime, pathlib.Path("export.mp3")
    )


if __name__ == "__main__":
    import asyncio

    asyncio.run(main())
